package com.amazon.kcp.library

import android.content.Context
import android.content.Intent
import android.widget.Toast
import com.amazon.kcp.application.*

import com.amazon.kcp.reader.ui.ISettingsControlModel
import com.amazon.kcp.reader.ui.SettingsControlClickableModel
import com.amazon.kcp.reader.ui.SettingsControlToggleModel
import com.amazon.kcp.reader.ui.SettingsControlToggleModel.ToggleState.ON
import com.amazon.kcp.reader.ui.SettingsControlToggleModel.ToggleState.OFF
import com.amazon.kcp.readingstreams.ReadingStreamUtil
import com.amazon.kcp.redding.ReddingActivity
import com.amazon.kcp.util.PackageUtils
import com.amazon.kcp.util.Utils
import com.amazon.kindle.download.DownloadBookNetworkUtils
import com.amazon.kindle.krx.readingstreams.IReadingStreamsEncoder
import com.amazon.kindle.krx.theme.Theme
import com.amazon.kindle.librarymodule.R
import com.amazon.kindle.sdcard.ExternalSDCardSettingToggleModel
import com.amazon.kindle.sdcard.ExternalSDCardUtils
import com.amazon.kindle.services.authentication.TokenKey

import java.util.regex.Pattern

/**
 * Class containing Standalone specific [ISettingsControlModel]s to show with the [SettingsFragment].
 * These models are to be added first to the [com.amazon.kcp.reader.ui.SettingsActivityControlRegister].
 * See [com.amazon.kcp.application.StandaloneKindleObjectFactory.getSettingsActivityControlRegister] for that.
 *
 * @author vasushar
 */
class StandaloneSettingsModels {

    companion object {
        /**
         * For a clickable button showing Device Name.
         */
        private class DeviceNameSettingButtonModel(val authManager: IAuthenticationManager,
                                                   val appCapabilities: AndroidApplicationCapabilities): SettingsControlClickableModel() {
            override fun onClick(activity: ReddingActivity) {
                // The rename device API does not work on Grover/Canary
                if (!(PackageUtils.hasGrover(activity) || PackageUtils.hasCanary(activity))) {
                    val intent = Intent(activity, EditDeviceNameActivity::class.java)
                    activity.startActivity(intent)
                }
            }

            override fun getTitle(context : Context): String = context.getString(R.string.device_name)

            override fun getSubTitle(context : Context): String? {
                return if (authManager.isAuthenticated) {
                    authManager.fetchToken(TokenKey.DEVICE_NAME)
                } else {
                    null
                }
            }

            override fun getCategory(context : Context): String = context.getString(R.string.your_device)
            override fun getId(): Int = R.id.settings_device_name
            override fun shouldCreateRow(context : Context): Boolean = appCapabilities.canChangeRegistrationSettings()
        }

        /**
         * For a clickable button to modify notification settings.
         */
        private class NotificationStatusSettingButtonModel(val factory: IKindleObjectFactory): SettingsControlClickableModel() {
            override fun onClick(activity: ReddingActivity) {
                val i = Intent(activity, NotificationsActivity::class.java)
                activity.startActivity(i)
            }

            override fun getTitle(context : Context): String = context.getString(R.string.notification_detail)

            override fun getSubTitle(context : Context): String {
                return context.resources.getString(R.string.notification_state,
                        if (factory.notificationSettingsManager.areNotificationsEnabled())
                            context.resources.getString(R.string.notifications_on)
                        else
                            context.resources.getString(R.string.notifications_off))
            }

            override fun getCategory(context : Context): String = context.getString(R.string.notifications_title)
            override fun getId(): Int = R.id.settings_notifications_status
            override fun shouldCreateRow(context : Context): Boolean = Utils.arePromotionsAllowed()
        }

        /**
         * Clickable button to deregister the device.
         */
        private class DeregisterSettingButtonModel(val authManager: IAuthenticationManager,
                                                   val appCapabilities: AndroidApplicationCapabilities): SettingsControlClickableModel() {
            // Regex pulled from https://w.amazon.com/index.php/India_Mobile/Risk_Mitigation_for_Indigo
            private val SURROGATE_EMAIL_REGEX = Pattern.compile("amzn1.account.\\w*@amazon-customer.([a-zA-Z]{2})")

            override fun onClick(activity: ReddingActivity) {
                // Implementation in {@link SettingsFragment#onSettingsItemClick()} since this requires
                // starting a new Fragment from the current Fragment.
            }

            override fun getTitle(context : Context): String = context.getString(R.string.logout_device)

            override fun getSubTitle(context : Context): String? {
                if (authManager.isAuthenticated) {
                    val emailAddress = authManager.fetchToken(TokenKey.USER_EMAIL)
                    if (!Utils.isNullOrEmpty(emailAddress) && !SURROGATE_EMAIL_REGEX.matcher(emailAddress).matches()) {
                        // Suppress display of the surrogate email address
                        return context.resources.getString(R.string.registered_to) + " " + emailAddress
                    }
                }
                return null
            }

            override fun getCategory(context : Context): String = context.getString(R.string.registration)
            override fun getId(): Int = R.id.settings_deregister_device
            override fun shouldCreateRow(context : Context): Boolean = appCapabilities.canChangeRegistrationSettings()
        }

        /**
         * Clickable button for showing details regarding Send To Kindle email.
         */
        private class SendToKindleSettingButtonModel(val authManager: IAuthenticationManager,
                                                     val appCapabilities: AndroidApplicationCapabilities): SettingsControlClickableModel() {
            override fun onClick(activity: ReddingActivity) {
                val intent = Intent(activity, SendToKindleDetailsActivity::class.java)
                activity.startActivity(intent)
            }

            override fun getTitle(context : Context): String = context.getString(R.string.whispersend_email_address)

            override fun getSubTitle(context : Context): String {
                return if (authManager.isAuthenticated) {
                    authManager.fetchToken(TokenKey.DEVICE_EMAIL)
                } else {
                    context.getString(R.string.not_logged_in)
                }
            }

            override fun getCategory(context : Context): String {
                return if (ExternalSDCardUtils.isExternalSdCardV2SupportEnabled()) {
                    context.getString(R.string.storage_setting_category)
                } else {
                    context.getString(R.string.personal_documents)
                }
            }
            override fun getId(): Int = R.id.settings_whispersend_email_address
            override fun shouldCreateRow(context : Context): Boolean = appCapabilities.canChangeRegistrationSettings()
                    && Utils.hasSendToKindleEmail()
        }

        /**
         * On/Off button to choose whether PDocs should be detected from storage.
         */
        private class SideloadedPDocSettingButtonModel(val userSettings: UserSettingsController): SettingsControlToggleModel() {
            override fun getCurrentState(activity: ReddingActivity): SettingsControlToggleModel.ToggleState {
                return if (userSettings.personalDocsSetting && activity.permissionsManager.hasExternalStoragePermission())
                    SettingsControlToggleModel.ToggleState.ON
                else
                    SettingsControlToggleModel.ToggleState.OFF
            }

            override fun onOnPressed(activity: ReddingActivity) {
                val onSuccess = Runnable { userSettings.personalDocsSetting = true }

                val onFailure = Runnable {
                    userSettings.personalDocsSetting = false
                    userSettings.haveRequestedPersonalDocsPermission = true
                    val toast = Toast.makeText(activity, activity.resources.getString(R.string.storage_permission_denied_toast), Toast.LENGTH_SHORT)
                    toast.show()
                }

                activity.permissionsManager.requestExternalStoragePermissions(onSuccess, onFailure)
            }

            override fun onOffPressed(activity: ReddingActivity) {
                userSettings.personalDocsSetting = false
            }

            override fun getTitle(context : Context): String = context.getString(R.string.storage_permission_setting_title)
            override fun getSubTitle(context : Context): String = context.getString(R.string.storage_permission_setting_detail)
            override fun getCategory(context : Context): String {
                return if (ExternalSDCardUtils.isExternalSdCardV2SupportEnabled()) {
                    context.getString(R.string.storage_setting_category)
                } else {
                    context.getString(R.string.personal_documents)
                }
            }
            override fun getId(): Int = R.id.settings_sideload_and_pdocs_enabled
            override fun shouldCreateRow(context : Context): Boolean = true
        }

        /**
         * On/Off button for toggling Whispersync setting.
         */
        private class WhispersyncSettingButtonModel(val userSettings: UserSettingsController): SettingsControlToggleModel() {
            override fun getCurrentState(activity: ReddingActivity): SettingsControlToggleModel.ToggleState {
                return if (userSettings.isAnnotationsSyncEnabled) ON else OFF
            }

            override fun onOnPressed(activity: ReddingActivity) {
                // Implementation in {@link SettingsFragment#onSettingsItemCheckBoxChanged()} because
                // it requires showing a fragment from existing fragment.
            }

            override fun onOffPressed(activity: ReddingActivity) {
                // Implementation in {@link SettingsFragment#onSettingsItemCheckBoxChanged()} because
                // it requires showing a fragment from existing fragment.
            }

            override fun getTitle(context : Context): String = context.getString(R.string.activity_data)
            override fun getSubTitle(context : Context): String = context.getString(R.string.activity_data_detail)
            override fun getCategory(context : Context): String = context.getString(R.string.activity_data_title)
            override fun getId(): Int = R.id.settings_activity_data_sync
            override fun shouldCreateRow(context : Context): Boolean = true
        }

        /**
         * Clickable button to modify current theme.
         */
        private class ThemeSettingButtonModel(val factory: IKindleObjectFactory,
                                              val encoder: IReadingStreamsEncoder) : SettingsControlClickableModel() {
            override fun onClick(activity: ReddingActivity) {
                val intent = Intent(activity, ColorThemeActivity::class.java)
                encoder.performAction(ReadingStreamUtil.CONTEXT_SETTINGS, ReadingStreamUtil.ACTION_OPEN_THEME_SETTINGS)
                activity.startActivity(intent)
            }

            override fun getTitle(context : Context): String = context.getString(R.string.theme_settings)

            override fun getSubTitle(context : Context): String {
                return when (factory.kindleReaderSDK.themeManager.theme) {
                    Theme.LIGHT -> context.getString(R.string.theme_light)
                    Theme.DARK -> context.getString(R.string.theme_dark)
                }
            }

            override fun getCategory(context : Context): String = context.getString(R.string.application_settings)
            override fun getId(): Int = R.id.settings_theme_settings
            override fun shouldCreateRow(context : Context): Boolean = factory.kindleReaderSDK.themeManager.areMultipleThemesSupported()
        }

        /**
         * On/Off button to modify Volume control setting.
         */
        private class VolumeSettingButtonModel(val userSettings: UserSettingsController,
                                               val encoder: IReadingStreamsEncoder) : SettingsControlToggleModel() {
            override fun getCurrentState(activity: ReddingActivity): SettingsControlToggleModel.ToggleState {
                encoder.recordSetting(ReadingStreamUtil.CONTEXT_READING, UserSettingsController.Setting.VOLUME_KEYS_PAGE_CONTROLS.name, userSettings.areVolumeKeysPageControls(), false)
                return if (userSettings.areVolumeKeysPageControls()) SettingsControlToggleModel.ToggleState.ON else SettingsControlToggleModel.ToggleState.OFF
            }

            override fun onOnPressed(activity: ReddingActivity) {
                userSettings.setVolumeKeysArePageControls(true)
                encoder.recordSetting(ReadingStreamUtil.CONTEXT_READING, UserSettingsController.Setting.VOLUME_KEYS_PAGE_CONTROLS.name, userSettings.areVolumeKeysPageControls(), true)
            }

            override fun onOffPressed(activity: ReddingActivity) {
                userSettings.setVolumeKeysArePageControls(false)
                encoder.recordSetting(ReadingStreamUtil.CONTEXT_READING, UserSettingsController.Setting.VOLUME_KEYS_PAGE_CONTROLS.name, userSettings.areVolumeKeysPageControls(), true)
            }

            override fun getTitle(context : Context): String = context.getString(R.string.volume_keys)
            override fun getSubTitle(context : Context): String = context.getString(R.string.volume_keys_detail)
            override fun getCategory(context : Context): String = context.getString(R.string.application_settings)
            override fun getId(): Int = R.id.settings_volume_keys
            override fun shouldCreateRow(context : Context): Boolean = true
        }

        /**
         * On/Of Setting to modify download setting in data network.
         */
        private class UseNetworkSettingButtonModel(val userSettings: UserSettingsController): SettingsControlToggleModel() {

            override fun getCurrentState(activity: ReddingActivity): SettingsControlToggleModel.ToggleState {
                return if (userSettings.downloadBookNetworkMode == DownloadBookNetworkUtils.DOWNLOAD_BOOK_NETWORK_MODE_ONLY_WIFI) {
                    SettingsControlToggleModel.ToggleState.OFF
                } else {
                    SettingsControlToggleModel.ToggleState.ON
                }
            }

            override fun onOnPressed(activity: ReddingActivity) {
                userSettings.downloadBookNetworkMode = DownloadBookNetworkUtils.DOWNLOAD_BOOK_NETWORK_MODE_USE_DATA
            }

            override fun onOffPressed(activity: ReddingActivity) {
                userSettings.downloadBookNetworkMode = DownloadBookNetworkUtils.DOWNLOAD_BOOK_NETWORK_MODE_ONLY_WIFI
            }

            override fun getTitle(context : Context): String = context.getString(R.string.use_data_network)
            override fun getSubTitle(context : Context): String = context.getString(R.string.use_data_network_detail)
            override fun getCategory(context : Context): String = context.getString(R.string.download_settings)
            override fun getId(): Int = R.id.settings_use_data_network
            override fun shouldCreateRow(context : Context): Boolean = context.resources.getBoolean(R.bool.show_network_settings)
        }

        /**
         * To get all the models specified in this class.
         *
         * @return a list of [ISettingsControlModel] objects that represent a setting buttons.
         */
        @JvmStatic
        fun getModels(factory: IKindleObjectFactory) : List<ISettingsControlModel> {

            val appCapabilities = factory.applicationCapabilities
            val authManager = factory.authenticationManager
            val userSettings = factory.userSettingsController
            val encoder = factory.kindleReaderSDK.readingStreamsEncoder

            // Note: the order of elements in this list matters.
            return listOf(DeviceNameSettingButtonModel(authManager, appCapabilities),
                    NotificationStatusSettingButtonModel(factory),
                    DeregisterSettingButtonModel(authManager, appCapabilities),
                    ExternalSDCardSettingToggleModel(factory.storageLocationPreference),
                    SendToKindleSettingButtonModel(authManager, appCapabilities),
                    SideloadedPDocSettingButtonModel(userSettings),
                    WhispersyncSettingButtonModel(userSettings),
                    ThemeSettingButtonModel(factory, encoder),
                    VolumeSettingButtonModel(userSettings, encoder),
                    UseNetworkSettingButtonModel(userSettings))
        }
    }
}
